/**
 * 
 */
package org.nasa.srfe;

import java.awt.Point;
import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.nasa.srfe.SaturnRingFeatureExtractor.SaturnRingFeatureInfo;

/**
 * <p>
 * Feature cross correlation class.
 * </p>
 * <p>
 * <b>Description:</b>
 * Feature information is categorized as same features including 
 * featureID and the number of same feature in the features.csv file.
 * </p>
 * @author Inwoo Chung (gutomitai@gmail.com)
 * @since Jun. 19, 2014
 *
 */
public class FeatureCrossCorrelation {

	/**
	 * Radius threshold.
	 */
	public final static double RADIUS_TH = 10.0; // Degree. 
	
	/**
	 * Longitude threshold.
	 */
	public final static double LONGITUDE_TH = 30.0; // km.
	
	/**
	 * Features file path.
	 */
	public final static String FEATURES_FILE_PATH = "features_cc.csv";
	
	/**
	 * Cross correlated feature class. 
	 */
	public static class CrossCorrelatedFeatureInfo extends SaturnRingFeatureInfo {
		
		/**
		 * Grouped feature ID.
		 */
		public int featureID = UNDETERMINED;
		/**
		 * The number of grouped features.
		 */
		public int count = UNDETERMINED;
		/**
		 * Flag for grouping.
		 */
		public boolean isUsed = false;
	}
	
	/**
	 * Constructor.
	 */
	public FeatureCrossCorrelation() {
	}

	/**
	 * Main.
	 * @param args
	 * 			features.csv file path name.
	 */
	public static void main(String[] args) {

		// Check exception
		if(args.length == 0) 
			throw new IllegalArgumentException("Can't extract cross correlated features " +
					"because there is no argument. " +
					"A features.csv file path name is necessary.");
		
		String featuresFilePath = args[0];
		
		// Extract cross correlated features.
		FeatureCrossCorrelation fcc = new FeatureCrossCorrelation();
		HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>> featuresByEachImage = null;
		
		try {
			 featuresByEachImage =
					fcc.extractCrossCorrelatedFeatures(featuresFilePath);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		// Save feature information including featureID, count.
		try {
			saveCSVFileForFeaturesIDCount(featuresByEachImage, FEATURES_FILE_PATH);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * <p>
	 * Save a csv file for features including featureID, count.
	 * </p>
	 * @author Inwoo Chung (gutomitai@gmail.com), supercharger(pratapk@live.com)
	 * @param features
	 * @param filePath
	 * @throws IOException 
	 */
	public static void saveCSVFileForFeaturesIDCount
		(HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>> 
			featuresByEachImage, String filePath) 
			throws IOException {
		
		// Check exception.
		if(featuresByEachImage == null || filePath == null) 
			throw new NullPointerException("Can't save a csv file. " +
					"because features by each image or filePath is null.");
		
		// Save.
		FileOutputStream out = new FileOutputStream(filePath);
		
		// Make feature header text.
		String featureLine = String.format("imageName uid centerPos.x " +
				"centerPos.y radius radiusMin radiusMax radiusStd " +
				"longitude longitudeMin longitudeMax intensityMax area " +
				"shortAxisLength longAxisLength elongation startTime meanMotion " +
				"featureID count\n");
		out.write(featureLine.getBytes());	
		
		Iterator<ArrayList<CrossCorrelatedFeatureInfo>> iterFeatures = 
				featuresByEachImage.values().iterator();
		ArrayList<CrossCorrelatedFeatureInfo> unitGroupFeatures = null;
		
		while(iterFeatures.hasNext()) {
			
			unitGroupFeatures = iterFeatures.next();
			
			Iterator<CrossCorrelatedFeatureInfo> iter = unitGroupFeatures.iterator();

			while(iter.hasNext()) {
				
				CrossCorrelatedFeatureInfo feature = iter.next();
				
				// Make feature line text.
				featureLine = String.format("%s %d %d %d %.12f %.12f " +
						"%.12f %.12f %.12f %.12f %.12f %e %.12f %.12f %.12f %.12f %s %.12f %d %d\n", 
						feature.imageName,
						feature.uid,
						feature.centerPos.x,
						feature.centerPos.y,
						feature.radius,
						feature.radiusMin,
						feature.radiusMax,
						feature.radiusStd,
						feature.longitude,
						feature.longitudeMin,
						feature.longitudeMax,
						feature.intensityMax,
						feature.area,
						feature.shortAxisLength,
						feature.longAxisLength,
						feature.elongation,
						feature.startTime,
						feature.meanMotion,
						feature.featureID,
						feature.count);
				
				out.write(featureLine.getBytes());
			}
		}
	
		out.flush();
		out.close();
	}
	
	/**
	 * Extract cross correlated features.
	 * @param featuresFilePath
	 * @return
	 * @throws IOException 
	 */
	public HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>> 
		extractCrossCorrelatedFeatures(String featuresFilePath)
				throws IOException {
		
		// Check exception.
		if(featuresFilePath == null) 
			throw new NullPointerException("Can't extract cross correlated features " +
					"because the features.csv file path name is null.");
		
		// Extract features for each image.
		HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>> featuresByEachImage = 
				extractFeaturesForEachImage(featuresFilePath);
		
		return featuresByEachImage;
	}

	/**
	 * Extract features for each image.
	 * @param features
	 * @return
	 * @throws IOException
	 */
	public HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>>
		extractFeaturesForEachImage(ArrayList<CrossCorrelatedFeatureInfo> features) throws IOException {
		
		// Check exception.
		if(features == null) 
			throw new NullPointerException("Can't extract cross correlated features " +
					"because the crossly correlated features instance is null.");
		
		// Categorize features by each image.
		HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>> featuresByEachImage = 
				categorizeFeaturesByEachImage(features);
		
		// Extract cross correlation for features.
		extractCrossCorrelationForFeatures(featuresByEachImage);
		
		return featuresByEachImage;
	}
	
	// Extract features for each image.
	private HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>>
		extractFeaturesForEachImage(String featuresFilePath) throws IOException {
		
		// Check exception.
		if(featuresFilePath == null) 
			throw new NullPointerException("Can't extract cross correlated features " +
					"because the features.csv file path name is null.");
		
		ArrayList<CrossCorrelatedFeatureInfo> features = 
				new ArrayList<CrossCorrelatedFeatureInfo>();
		
		// Get input stream for feature information.
		BufferedReader in = new BufferedReader(new FileReader(featuresFilePath));
		
		// Eliminate header.
		String lineString = in.readLine(); 
		
		lineString = in.readLine();
		String delimiter = " ";
		String[] infos = null;
		int index = 0;
		
		// Parse.
		// Get features.
		while(lineString != null) {
			
			// Parse information.
			infos = lineString.split(delimiter);
			CrossCorrelatedFeatureInfo feature = new CrossCorrelatedFeatureInfo();
			feature.centerPos = new Point();
			feature.region = new Rectangle();
			
			feature.imageName = infos[index++];
			feature.uid = Integer.parseInt(infos[index++]);
			feature.centerPos.x = Integer.parseInt(infos[index++]);
			feature.centerPos.y = Integer.parseInt(infos[index++]);
			feature.radius = Double.parseDouble(infos[index++]);
			feature.radiusMin = Double.parseDouble(infos[index++]);
			feature.radiusMax = Double.parseDouble(infos[index++]);
			feature.radiusStd = Double.parseDouble(infos[index++]);
			feature.longitude = Double.parseDouble(infos[index++]);
			feature.longitudeMin = Double.parseDouble(infos[index++]);
			feature.longitudeMax = Double.parseDouble(infos[index++]);
			feature.intensityMax = Double.parseDouble(infos[index++]);
			feature.area = Double.parseDouble(infos[index++]);
			feature.shortAxisLength = Double.parseDouble(infos[index++]);
			feature.longAxisLength = Double.parseDouble(infos[index++]);
			feature.elongation = Double.parseDouble(infos[index++]);
			feature.startTime = infos[index++];
			feature.startTimeMS = Double.parseDouble(infos[index++]);
			feature.meanMotion = Double.parseDouble(infos[index++]);

			features.add(feature);
			
			lineString = in.readLine();
			index = 0;
		}
		
		in.close();
		
		// Categorize features by each image.
		HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>> featuresByEachImage = 
				categorizeFeaturesByEachImage(features);
		
		// Extract cross correlation for features.
		extractCrossCorrelationForFeatures(featuresByEachImage);
		
		return featuresByEachImage;
	}
	
	// Extract cross correlation for features.
	private void extractCrossCorrelationForFeatures
		(HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>> featuresByEachImage) {
		
		// Check exception.
		if(featuresByEachImage == null) 
			throw new NullPointerException("Can't extract cross correlation for features " +
					"because an each image features instance is null.");
		
		// Extract grouped features.
		HashMap<Integer, HashSet<CrossCorrelatedFeatureInfo>> groupedFeatures = 
				new HashMap<Integer, HashSet<CrossCorrelatedFeatureInfo>>(); 
		int featureID = 0; // Zero based.
		Set<String> imageNameSet = featuresByEachImage.keySet();
		int numImages = imageNameSet.size();
		int index = 0;
		
		CrossCorrelatedFeatureInfo[][] featuresArrayByEachImage = 
				new CrossCorrelatedFeatureInfo[numImages][];
		
		Iterator<String> iterImageName = imageNameSet.iterator();
		
		while(iterImageName.hasNext()) {
			
			featuresArrayByEachImage[index++] = 
					featuresByEachImage.get(iterImageName.next()).
					toArray(new CrossCorrelatedFeatureInfo[0]);
		}
		
		// Group features with same revolution property.
		double refLongitude = 0.0;
		double compLongitude = 0.0;
		double longDiff = 0.0;
		double radiusDiff = 0.0;
		
		for(int i = 0; i <= numImages - 1; i++) {
			for(int k = 0; k <= featuresArrayByEachImage[i].length - 1; k++) {
				
				if(featuresArrayByEachImage[i][k].isUsed == true) 
					continue;
				
				HashSet<CrossCorrelatedFeatureInfo> unitGroupedFeatures = 
						new HashSet<CrossCorrelatedFeatureInfo>();
				
				unitGroupedFeatures.add(featuresArrayByEachImage[i][k]);
				featuresArrayByEachImage[i][k].isUsed = true;

				for(int j = i + 1; j <= numImages - 1; j++) {
					for(int m = 0; m <= featuresArrayByEachImage[j].length - 1; m++) {
						
						// Check whether there is a same revolution property.
						// Check whether a component was grouped.
						if(featuresArrayByEachImage[j][m].isUsed) 
							continue;
						
						// Check radius.
						radiusDiff = 
								Math.abs(featuresArrayByEachImage[i][k].radius - 
										featuresArrayByEachImage[j][m].radius);
						
						if(radiusDiff > RADIUS_TH) 
							continue;
						
						compLongitude = featuresArrayByEachImage[j][m].longitude;
						refLongitude = ((featuresArrayByEachImage[j][m].startTimeMS - 
								featuresArrayByEachImage[i][k].startTimeMS)/1000.0*
								featuresArrayByEachImage[i][k].meanMotion +
								featuresArrayByEachImage[i][k].longitude)%360;
						
						longDiff = Math.abs( compLongitude - refLongitude);
						
						// Check longitude difference.
						if(longDiff > LONGITUDE_TH) 
							continue;
						
						// Add a component into the group.
						unitGroupedFeatures.add(featuresArrayByEachImage[j][m]);
						featuresArrayByEachImage[j][m].isUsed = true;
					}	
				}
				
				groupedFeatures.put(featureID++, unitGroupedFeatures);
			}
		}
		
		// Add featureID, count.
		Set<Integer> featureIDs = groupedFeatures.keySet();
		int numFeatureIDs = featureIDs.size();
		int featureIDKey = 0;
		int sizeUnitGroup = 0;
		HashSet<CrossCorrelatedFeatureInfo> unitGroupedFeatures = null;
		CrossCorrelatedFeatureInfo feature = null;
		
		Iterator<Integer> iterFeatureIDs = featureIDs.iterator();
		
		while(iterFeatureIDs.hasNext()) {
			
			featureIDKey = iterFeatureIDs.next();
			unitGroupedFeatures = groupedFeatures.get(featureIDKey);
			sizeUnitGroup = unitGroupedFeatures.size();
			
			Iterator<CrossCorrelatedFeatureInfo> iterFeature = 
					unitGroupedFeatures.iterator();
			
			while(iterFeature.hasNext()) {
				
				feature = iterFeature.next();
				feature.featureID = featureIDKey;
				feature.count = sizeUnitGroup;
			}
		}
	}
	
	// Categorize features by each image.
	private HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>> 
		categorizeFeaturesByEachImage(ArrayList<CrossCorrelatedFeatureInfo> features) {
		
		// Check exception.
		if(features == null) 
			throw new NullPointerException("Can't categorize features by each image " +
					"because an input features instance is null.");
		
		// Categorize.
		// Get the kind of images.
		HashSet<String> imageNames = new HashSet<String>();
		CrossCorrelatedFeatureInfo info = null;
		
		Iterator<CrossCorrelatedFeatureInfo> iterFeatures = features.iterator();
		
		while(iterFeatures.hasNext()) {
			
			info = iterFeatures.next();
			
			imageNames.add(info.imageName);
		}
		
		// Create features by each image.
		HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>> 
			featuresByEachImage = new HashMap<String, ArrayList<CrossCorrelatedFeatureInfo>>();
		
		Iterator<String> iterImageNames = imageNames.iterator();
		String imageName = null;
		ArrayList<CrossCorrelatedFeatureInfo> unitFeatures = null;
		
		while(iterImageNames.hasNext()) {
			
			imageName = iterImageNames.next();
			unitFeatures = new ArrayList<CrossCorrelatedFeatureInfo>();
			
			// Extract components with a same image name.
			iterFeatures = features.iterator();
			
			while(iterFeatures.hasNext()) {
				
				info = iterFeatures.next();
				
				if(info.imageName.equals(imageName)) 
					unitFeatures.add(info);
			}
			
			featuresByEachImage.put(imageName, unitFeatures );
		}
		
		return featuresByEachImage;
	}
	
	/**
	 * Parse UTC time.
	 * @param UTCTime
	 * @return
	 * @throws NumberFormatException
	 */
	public static long parseUTCTime(String UTCTime) throws 
		NumberFormatException {
		
		// Check exception.
		if(UTCTime == null) 
			throw new NullPointerException("Can't parse UTC time " +
					"because an input UTC time string is null.");
	
		// Get UTC time from A.D zero in millisecond unit.
		Calendar cal = Calendar.getInstance();
		
		// Parse.
		// Year.
		String[] result = UTCTime.split("-");
		
		// Check exception.
		if(result.length != 2) 
			throw new IllegalArgumentException("Can't parse UTC time " +
					"because an input UTC time string isn't valid.");
		
		cal.set(Calendar.YEAR, Integer.parseInt(result[0]));
		
		// Days.
		result = result[1].split("T");
		
		// Check exception.
		if(result.length != 2) 
			throw new IllegalArgumentException("Can't parse UTC time " +
					"because an input UTC time string isn't valid.");
		
		cal.set(Calendar.DAY_OF_YEAR, Integer.parseInt(result[0]));
		
		// Hour, minute.
		result = result[1].split(":");
		
		// Check exception.
		if(result.length != 3) 
			throw new IllegalArgumentException("Can't parse UTC time " +
							"because an input UTC time string isn't valid.");
		
		cal.set(Calendar.HOUR, Integer.parseInt(result[0]));
		cal.set(Calendar.MINUTE, Integer.parseInt(result[1]));
		
		// Second, millisecond.
		double secDouble = Double.parseDouble(result[2]);
		int sec = (int)secDouble;
		int milliSec = (int)((secDouble - sec)*1000.0);
		
		cal.set(Calendar.SECOND, sec);
		cal.set(Calendar.MILLISECOND, milliSec);
		
		return cal.getTimeInMillis();
	}
}
